/*!
 * Copyright (c) Jayly. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */
import { ScoreboardIdentityType, system, world } from "@minecraft/server";
const version = "1.0.8";
const str = () => ('00000000000000000' + (Math.random() * 0xffffffffffffffff).toString(16)).slice(-16);
const uuid = () => {
    const [a, b] = [str(), str()];
    return `${a.slice(0, 8)}-${a.slice(8, 12)}-4${a.slice(13)}-a${b.slice(1, 4)}-${b.slice(4)}`;
};
const allowedTypes = ["string", "number", "boolean"];
const encrypt = (data, salt) => {
    const encryptedChars = [];
    for (let i = 0; i < data.length; i++) {
        const charCode = data.charCodeAt(i) + salt.charCodeAt(i % salt.length);
        encryptedChars.push(charCode);
    }
    return String.fromCharCode(...encryptedChars);
};
const decrypt = (encrypted, salt) => {
    const decryptedChars = [];
    for (let i = 0; i < encrypted.length; i++) {
        const charCode = encrypted.charCodeAt(i) - salt.charCodeAt(i % salt.length);
        decryptedChars.push(charCode);
    }
    return String.fromCharCode(...decryptedChars);
};
const CreateCrashReport = (action, data, error, salt) => {
    console.warn("[JaylyDB] Failed to " + action + " JSON data.", "\nVersion: " + version, "\nData: " + data, "\nSalt: " + salt, "\nError: " + error.message, "\n" + error.stack);
    throw new Error(`Failed to ${action} data. Please check content log file for more info.\n`);
};
const DisplayName = {
    parse(text, salt) {
        try {
            const a = JSON.parse(`"${salt ? decrypt(text, salt) : text}"`);
            return JSON.parse(`{${a}}`);
        }
        catch (error) {
            CreateCrashReport("load", text, error, salt);
        }
    },
    stringify(value, salt) {
        try {
            const a = JSON.stringify(JSON.stringify(value).slice(1, -1)).slice(1, -1);
            return salt ? encrypt(a, salt) : a;
        }
        catch (error) {
            CreateCrashReport("save", JSON.stringify(value), error, salt);
        }
    }
};
const overworld = world.getDimension("overworld");
class JaylyDB {
    objective;
    encrypted;
    participants = new Map();
    localCache = new Map();
    salt;
    warningSent = false;
    SYNC_OK = true;
    tempCache = new Map();
    updateParticipants(fetchCache = false) {
        const id = this.objective.id.substring(this.objective.id.indexOf(":") + 1);
        if (this.tempCache.size <= 0 && this.warningSent === true) {
            this.warningSent = false;
        }
        else if (this.tempCache.size > 0 && this.warningSent === false) {
            this.warningSent = true;
        }
        ;
        try {
            for (const [key, value] of this.tempCache.entries()) {
                overworld.runCommandAsync(`scoreboard players set "${value}" ${this.objective.id} 0`);
                this.tempCache.delete(key);
            }
            ;
        }
        catch (error) {
            if (error.message !== "Runtime failure, command queue is full.")
                throw error;
        }
        ;
        this.participants.clear();
        if (fetchCache)
            this.localCache.clear();
        for (const participant of this.objective.getParticipants()) {
            if (participant.type !== ScoreboardIdentityType.fakePlayer)
                continue;
            const data = DisplayName.parse(participant.displayName, this.salt);
            const key = Object.keys(data)[0];
            const value = data[key];
            this.participants.set(key, participant);
            if (fetchCache)
                this.localCache.set(key, value);
        }
        ;
        if (this.SYNC_OK === false) {
            console.warn(`[JaylyDB] Database '${id}' is now sync with disk.`);
            this.SYNC_OK = true;
        }
        ;
    }
    constructor(id, encrypted = false) {
        this.objective = world.scoreboard.getObjective("jaylydb:" + id) ?? world.scoreboard.addObjective("jaylydb:" + id, uuid());
        this.encrypted = encrypted;
        this.salt = this.encrypted ? this.objective.displayName : undefined;
        this.updateParticipants(true);
        system.runInterval(() => {
            if (!!world.scoreboard.getObjective("jaylydb:" + id))
                return this.updateParticipants();
            else if (this.SYNC_OK === true)
                console.error(`[JaylyDB] There is a sync issue with database '${id}'.`);
            this.localCache.forEach((value, key) => {
                const encoded = DisplayName.stringify({ [key]: value }, this.salt);
                this.tempCache.set(key, encoded);
            });
            this.SYNC_OK = false;
        });
    }
    get size() {
        return this.participants.size;
    }
    clear() {
        this.participants.forEach(participant => this.objective.removeParticipant(participant));
        this.updateParticipants();
    }
    delete(key) {
        const participant = this.participants.get(key);
        if (!participant)
            return false;
        const success = this.objective.removeParticipant(participant);
        this.participants.delete(key);
        this.localCache.delete(key);
        return success;
    }
    forEach(callbackfn) {
        for (const [key, value] of this.entries())
            callbackfn(value, key, this);
    }
    get(key) {
        if (this.localCache.has(key))
            return this.localCache.get(key);
        const participant = this.participants.get(key);
        if (!participant)
            return undefined;
        const displayName = participant.displayName;
        const data = DisplayName.parse(displayName, this.salt);
        this.localCache.set(key, data[key]);
        return data[key];
    }
    has(key) {
        return this.localCache.has(key);
    }
    set(key, value) {
        if (!allowedTypes.includes(typeof value))
            throw new TypeError("JaylyDB::set only accepts a value of string, number, or boolean.");
        const encoded = DisplayName.stringify({ [key]: value }, this.salt);
        if (encoded.length > 32767)
            throw new RangeError("JaylyDB::set only accepts a string value less than 32767 characters.");
        if (this.localCache.get(key) === value)
            return this;
        const participant = this.participants.get(key);
        this.participants.delete(key);
        if (participant)
            this.objective.removeParticipant(participant);
        this.tempCache.set(key, encoded);
        this.localCache.set(key, value);
        return this;
    }
    *entries() {
        for (const iterator of this.localCache.entries())
            yield iterator;
    }
    *keys() {
        for (const [key] of this.entries())
            yield key;
    }
    *values() {
        for (const [, value] of this.entries())
            yield value;
    }
    [Symbol.iterator]() {
        return this.entries();
    }
    [Symbol.toStringTag] = JaylyDB.name;
}
;
export { JaylyDB };
